using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace USBDevCtrl
{
	class HIDDevice
	{
	IntPtr devNotifyHandle = IntPtr.Zero;

	// Enum values from HIDP_REPORT_TYPE
	public const short HidP_Input = 0;
	public const short HidP_Output = 1;
	public const short HidP_Feature = 2;

	[StructLayout(LayoutKind.Sequential)]
	public struct HIDD_ATTRIBUTES
	{
		public int Size;
		public short VendorID;
		public short ProductID;
		public short VersionNumber;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct HIDP_CAPS
	{
		public short Usage;
		public short UsagePage;
		public short InputReportByteLength;
		public short OutputReportByteLength;
		public short FeatureReportByteLength;
		[MarshalAs(UnmanagedType.ByValArray, SizeConst = 17)]
		public short[] Reserved;
		public short NumberLinkCollectionNodes;
		public short NumberInputButtonCaps;
		public short NumberInputValueCaps;
		public short NumberInputDataIndices;
		public short NumberOutputButtonCaps;
		public short NumberOutputValueCaps;
		public short NumberOutputDataIndices;
		public short NumberFeatureButtonCaps;
		public short NumberFeatureValueCaps;
		public short NumberFeatureDataIndices;

	}

	[StructLayout(LayoutKind.Sequential)]
	public struct HIDP_BUTTON_CAPS
	{
		public short UsagePage;
		public byte  ReportID;
		public int  IsAlias;

		public short BitField;
		public short LinkCollection;   // A unique internal index pointer

		public short LinkUsage;
		public short LinkUsagePage;

		public int IsRange;
		public int IsStringRange;
		public int IsDesignatorRange;
		public int IsAbsolute;

			// 10 reserved fields
		[MarshalAs(UnmanagedType.ByValArray, SizeConst = 10)]
		public uint[] Reserved;
		public short  UsageMin,      UsageMax;
        public short  StringMin,     StringMax;
        public short  DesignatorMin, DesignatorMax;
        public short  DataIndexMin,  DataIndexMax;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct HIDP_VALUE_CAPS
	{
		public short UsagePage;
		public byte ReportID;
		public int IsAlias;
		public short BitField;
		public short LinkCollection;
		public short LinkUsage;
		public short LinkUsagePage;
		public int IsRange;
		public int IsStringRange;
		public int IsDesignatorRange;
		public int IsAbsolute;
		public int HasNull;
		public byte Reserved;
		public short BitSize;
		public short ReportCount;
		public short Reserved2;
		public short Reserved3;
		public short Reserved4;
		public short Reserved5;
		public short Reserved6;
		public int LogicalMin;
		public int LogicalMax;
		public int PhysicalMin;
		public int PhysicalMax;
		public short UsageMin;
		public short UsageMax;
		public short StringMin;
		public short StringMax;
		public short DesignatorMin;
		public short DesignatorMax;
		public short DataIndexMin;
		public short DataIndexMax;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct HIDP_DATA
	{
		public short DataIndex;
		public short Reserved;
		public uint RawValue;
	}

	[DllImport("hid.dll")]
		static public extern bool HidD_FlushQueue(IntPtr HidDeviceObject);

	[DllImport("hid.dll")]
	static public extern bool HidD_FreePreparsedData(IntPtr PreparsedData);

	[DllImport("hid.dll")]
		static public extern int HidD_GetAttributes(IntPtr HidDeviceObject, ref HIDD_ATTRIBUTES Attributes);

	[DllImport("hid.dll")]
		static public extern bool HidD_GetFeature(IntPtr HidDeviceObject, [In] Byte[] lpReportBuffer, int ReportBufferLength);

	[DllImport("hid.dll")]
		static public extern bool HidD_GetInputReport(IntPtr HidDeviceObject, [In] Byte[] lpReportBuffer, int ReportBufferLength);

	[DllImport("hid.dll")]
	static public extern void HidD_GetHidGuid(ref System.Guid HidGuid);

	[DllImport("hid.dll")]
		static public extern bool HidD_GetNumInputBuffers(IntPtr HidDeviceObject, ref int NumberBuffers);

	[DllImport("hid.dll")]
		static public extern bool HidD_GetPreparsedData(IntPtr HidDeviceObject, ref IntPtr PreparsedData);

	[DllImport("hid.dll")]
	static public extern bool HidD_SetFeature(IntPtr HidDeviceObject, [Out] Byte[]  lpReportBuffer, int ReportBufferLength);

	[DllImport("hid.dll")]
		static public extern bool HidD_SetNumInputBuffers(IntPtr HidDeviceObject, int NumberBuffers);

	[DllImport("hid.dll")]
		static public extern bool HidD_SetOutputReport(IntPtr HidDeviceObject, [Out] Byte[] lpReportBuffer, int ReportBufferLength);

	[DllImport("hid.dll")]
	static public extern int HidP_GetCaps(IntPtr PreparsedData, ref HIDP_CAPS Capabilities);

	[DllImport("hid.dll")]
	static public extern int HidP_GetValueCaps(short ReportType, ref byte ValueCaps, ref short ValueCapsLength, IntPtr PreparsedData);

	// API declarations relating to device management (SetupDixxx and
	// RegisterDeviceNotification functions).

	// ******************************************************************************
	// API constants
	// ******************************************************************************

	// from setupapi.h
	public const short DIGCF_PRESENT = 0x00000002;
	public const short DIGCF_DEVICEINTERFACE = 0x00000010;

	// ******************************************************************************
	// Structures and classes for API calls, listed alphabetically
	// ******************************************************************************

	// From winuser.h
		public const int WM_DEVICECHANGE = 0x0219;

		public const int DEVICE_NOTIFY_WINDOW_HANDLE = 0x00000000;
		public const int DEVICE_NOTIFY_SERVICE_HANDLE = 0x00000001;
		public const int DEVICE_NOTIFY_ALL_INTERFACE_CLASSES = 0x00000004;

	// Definitions from dbt.h
		public const int DBT_DEVTYP_OEM = 0x00000000;  // oem-defined device type
		public const int DBT_DEVTYP_DEVNODE = 0x00000001;  // devnode number
		public const int DBT_DEVTYP_VOLUME = 0x00000002;  // logical volume
		public const int DBT_DEVTYP_PORT = 0x00000003;  // serial, parallel
		public const int DBT_DEVTYP_NET = 0x00000004;  // network resource
		public const int DBT_DEVTYP_DEVICEINTERFACE = 0x00000005;  // device interface class
		public const int DBT_DEVTYP_HANDLE = 0x00000006;  // file system handle

		public const int DBT_DEVICEARRIVAL = 0x8000;  // system detected a new device
		public const int DBT_DEVICEQUERYREMOVE = 0x8001;  // wants to remove, may fail
		public const int DBT_DEVICEQUERYREMOVEFAILED = 0x8002;  // removal aborted
		public const int DBT_DEVICEREMOVEPENDING = 0x8003;  // about to remove, still avail.
		public const int DBT_DEVICEREMOVECOMPLETE = 0x8004;  // device is gone

	[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
	public class DEV_BROADCAST_DEVICEINTERFACE
	{
		public int dbcc_size;
		public int dbcc_devicetype;
		public int dbcc_reserved;
		[MarshalAs(UnmanagedType.ByValArray, ArraySubType = UnmanagedType.U1, SizeConst = 16)]
		public byte[] dbcc_classguid;
		[MarshalAs(UnmanagedType.ByValArray, SizeConst = 255)]
		public char[] dbcc_name;
	}

	[StructLayout(LayoutKind.Sequential)]
	public class  DEV_BROADCAST_VOLUME
	{
		public int dbcv_size;
		public int dbcv_devicetype;
		public int dbcv_reserved;
		public int dbcv_unitmask;
		public short dbcv_flags;
	}

	[StructLayout(LayoutKind.Sequential)]
	public class DEV_BROADCAST_PORT
	{
		public int dbcp_size;
		public int dbcp_devicetype;
		public int dbcp_reserved;
		[MarshalAs(UnmanagedType.ByValArray, SizeConst = 255)]
		public char[] dbcp_name;
	}

	[StructLayout(LayoutKind.Sequential)]
	public class DEV_BROADCAST_HANDLE
	{
		public int dbch_size;
		public int dbch_devicetype;
		public int dbch_reserved;
		public int dbch_handle;
		public int dbch_hdevnotify;
	}

	[StructLayout(LayoutKind.Sequential)]
	public class DEV_BROADCAST_HDR
	{
		public int dbch_size;
		public int dbch_devicetype;
		public int dbch_reserved;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct SP_DEVICE_INTERFACE_DATA
	{
		public int cbSize;
		public System.Guid InterfaceClassGuid;
		public int Flags;
		public int Reserved;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct SP_DEVICE_INTERFACE_DETAIL_DATA
	{
		public int cbSize;
		public string DevicePath;
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct SP_DEVINFO_DATA
	{
		public int cbSize;
		public System.Guid ClassGuid;
		public int DevInst;
		public int Reserved;
	}

	// ******************************************************************************
	// API functions, listed alphabetically
	// ******************************************************************************
	[DllImport("user32.dll", CharSet = CharSet.Auto)]
	static public extern IntPtr RegisterDeviceNotification(IntPtr hRecipient, IntPtr NotificationFilter, int Flags);

	[DllImport("setupapi.dll")]
	static public extern int SetupDiCreateDeviceInfoList(ref System.Guid ClassGuid, int hwndParent);

	[DllImport("setupapi.dll")]
	static public extern int SetupDiDestroyDeviceInfoList(IntPtr DeviceInfoSet);

	[DllImport("setupapi.dll")]
	static public extern int SetupDiEnumDeviceInterfaces(IntPtr DeviceInfoSet, int DeviceInfoData, ref System.Guid InterfaceClassGuid, int MemberIndex, ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData);

	[DllImport("setupapi.dll", CharSet = CharSet.Auto)]
	static public extern IntPtr SetupDiGetClassDevs(ref System.Guid ClassGuid, string Enumerator, int hwndParent, int Flags);

	[DllImport("setupapi.dll", CharSet = CharSet.Auto)]
	static public extern bool SetupDiGetDeviceInterfaceDetail(IntPtr DeviceInfoSet, ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData, IntPtr DeviceInterfaceDetailData, int DeviceInterfaceDetailDataSize, ref int RequiredSize, IntPtr DeviceInfoData);

	[DllImport("user32.dll")]
	static public extern bool UnregisterDeviceNotification(IntPtr Handle);
		
		public bool FillDeviceInfo()
		{
			InputDataLength = Caps.InputReportByteLength;
			InputReport = new byte[Caps.InputReportByteLength];
			OutputDataLength = Caps.OutputReportByteLength;
			OutputReport = new byte[Caps.OutputReportByteLength];
			FeatureDataLength = Caps.FeatureReportByteLength;
			FeatureReport = new byte[Caps.FeatureReportByteLength];

			return true;
		}

		public bool OpenHidDevice(string inDevicePath)
		{
			uint accessFlags = 0;
			uint sharingFlags = 0;
			bool bSuccess;

			DevicePath = inDevicePath;

			if (DevicePath == null)
				// No device name
				return false;

			accessFlags = Win32Comm.GENERIC_READ | Win32Comm.GENERIC_WRITE;
			sharingFlags = Win32Comm.FILE_SHARE_READ | Win32Comm.FILE_SHARE_WRITE;

			HidDevice = Win32Comm.CreateFile(DevicePath, accessFlags, sharingFlags,
							IntPtr.Zero, Win32Comm.OPEN_EXISTING, 0, IntPtr.Zero);

			if (HidDevice.ToInt32() == Win32Comm.INVALID_HANDLE_VALUE)
			{
				DeviceDetected = false;
				DevicePath = null;
				return false;
			}

			ReadHandle = Win32Comm.CreateFile(DevicePath, accessFlags, sharingFlags,
							IntPtr.Zero, Win32Comm.OPEN_EXISTING,
							Win32Comm.FILE_FLAG_OVERLAPPED, IntPtr.Zero);

			if (ReadHandle.ToInt32() == Win32Comm.INVALID_HANDLE_VALUE)
			{
				DeviceDetected = false;
				DevicePath = null;
				return false;
			}
			HIDOverlapped.hEvent = IntPtr.Zero;

			// Fill in the reset of the HidDevice structure
			if (!HidD_GetPreparsedData(HidDevice, ref Ppd))
			{
				DeviceDetected = false;
				DevicePath = null;
				Win32Comm.CloseHandle(HidDevice);
				Win32Comm.CloseHandle(ReadHandle);
				return false;
			}

			if (HidD_GetAttributes(HidDevice, ref Attributes) == 0)
			{
				DeviceDetected = false;
				DevicePath = null;
				Win32Comm.CloseHandle(HidDevice);
				Win32Comm.CloseHandle(ReadHandle);
				HidD_FreePreparsedData(Ppd);

				return false;
			}

			if (HidP_GetCaps(Ppd, ref Caps) == 0)
			{
				DeviceDetected = false;
				DevicePath = null;
				Win32Comm.CloseHandle(HidDevice);
				Win32Comm.CloseHandle(ReadHandle);
				HidD_FreePreparsedData(Ppd);

				return false;
			}

			int numBuffers = 0;
			if (HidD_GetNumInputBuffers(HidDevice, ref numBuffers))
			{
				int newNumBufs = numBuffers;
			}

			bSuccess = FillDeviceInfo();

			return bSuccess;
		}

		public void CloseHidDevice(bool FreeDeviceInfo)
		{
			DeviceDetected = false;

			DevicePath = null;

			if (HidDevice.ToInt32() != Win32Comm.INVALID_HANDLE_VALUE)
				Win32Comm.CloseHandle(HidDevice);
		}

		public bool FindHID(uint VendorID, uint ProductID)
		{
			//Use a series of API calls to find a HID with a matching Vendor and Product ID.
			SP_DEVICE_INTERFACE_DATA devInfoData = new SP_DEVICE_INTERFACE_DATA();
			int index = 0;
			bool foundDevice = false;
			int Result;
			SP_DEVICE_INTERFACE_DETAIL_DATA detailData;
			IntPtr hDevInfo = IntPtr.Zero;
			System.Guid HidGuid = System.Guid.NewGuid();
			int Required = 0;
			int Length = 0;

			// Get the GUID for HIDs
			HidD_GetHidGuid(ref HidGuid);

			// Get a handle to a device information set for all installed devices.
			hDevInfo = SetupDiGetClassDevs(ref HidGuid, null, 0,
				DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

			devInfoData.cbSize = Marshal.SizeOf(devInfoData);

			for (index = 0; ; index++)
			{
				// Get the next device in the list
				Result = SetupDiEnumDeviceInterfaces(hDevInfo, 0, ref HidGuid,
					index, ref devInfoData);

				if (Result != 0)
				{
					// Found a HID - now see if it is the right one

					// Determine how much data is needed to represent the
					// device detail information
					SetupDiGetDeviceInterfaceDetail(hDevInfo, ref devInfoData,
						IntPtr.Zero, 0, ref Length, IntPtr.Zero);

					//Allocate memory for the hDevInfo structure, using the returned Length.
					detailData = new SP_DEVICE_INTERFACE_DETAIL_DATA();
					detailData.cbSize = System.Runtime.InteropServices.Marshal.SizeOf(devInfoData);

					// Allocate memory for the MyDeviceInterfaceDetailData Structure using the returned buffer size.
					IntPtr detailDataBuf = Marshal.AllocHGlobal(Length);

					// Store cbSize in the first 4 bytes of the array
					Marshal.WriteInt32(detailDataBuf, 4 + Marshal.SystemDefaultCharSize);

					// Retrieve the actual device information.
					SetupDiGetDeviceInterfaceDetail(hDevInfo, ref devInfoData,
						detailDataBuf, Length, ref Required, IntPtr.Zero);

					// Skip over cbsize (4 bytes) to get the address of the devicePathName.
					IntPtr pdevicePathName = new IntPtr(detailDataBuf.ToInt32() + 4);

					// Open the device for read/write
					OpenHidDevice(Marshal.PtrToStringAuto(pdevicePathName));

					//Is it the desired device?
					foundDevice = false;

					if ((Attributes.VendorID == VendorID)
						&& (Attributes.ProductID == ProductID))
					{
						//Both the Product and Vendor IDs match.
						foundDevice = true;

						// Show the device's capablities.
						if (HIDOverlapped.hEvent.ToInt32() == 0)
						{
							HIDOverlapped.hEvent = Win32Event.CreateEvent(0, 1, 1, ""); ;
							HIDOverlapped.Offset = 0;
							HIDOverlapped.OffsetHigh = 0;
						}
						break;
					}
					else
						//The Vendor ID doesn't match.
						CloseHidDevice(true);

					//Free the memory used by the detailData structure (no longer needed).
					Marshal.FreeHGlobal(detailDataBuf);
				}
				else
					// No more HID devices in the list
					break;
			}

			//Free the memory reserved for hDevInfo by SetupDiClassDevs.
			SetupDiDestroyDeviceInfoList(hDevInfo);

			DeviceDetected = foundDevice;

			return DeviceDetected;
		}

		public bool RegisterForDeviceNotifications(IntPtr formHandle, Guid classGuid)
		{
			// A DEV_BROADCAST_DEVICEINTERFACE header holds information about the request.
			DEV_BROADCAST_DEVICEINTERFACE devIF = new DEV_BROADCAST_DEVICEINTERFACE();
			IntPtr devIFBuffer;

			int size;
			// Set up DEV_BROADCAST_DEVICEINTERFACE
			size = Marshal.SizeOf(devIF);
			devIF.dbcc_size = size;
			devIF.dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE;
			devIF.dbcc_reserved = 0;
			devIF.dbcc_classguid = classGuid.ToByteArray();

			// Allocate unmanaged memory to hold a copy of the structure
			devIFBuffer = Marshal.AllocHGlobal(size);

			// Copy managed structure into unmanaged memory
			Marshal.StructureToPtr(devIF, devIFBuffer, true);

			// Call Windows API to register for notifications
			devNotifyHandle = RegisterDeviceNotification(formHandle, devIFBuffer, DEVICE_NOTIFY_WINDOW_HANDLE);

			// Marshal data from the unmanaged block DevBroadcastDeviceInterfaceBuffer to
			// the managed object DevBroadcastDeviceInterface
			Marshal.PtrToStructure(devIFBuffer, devIF);

			// Free the unmanaged memory
			Marshal.FreeHGlobal(devIFBuffer);

			// Find out if RegisterDeviceNotification was successful.
			if (devNotifyHandle == IntPtr.Zero)
				return false;

			return true;
		}

		// Properties
		public bool DeviceDetected;
		public string  DevicePath;
		public IntPtr HidDevice; // A file handle to the hid device.
		public IntPtr ReadHandle; // File handle for overlapped I/O
		public Win32Comm.OVERLAPPED HIDOverlapped;

		public IntPtr Ppd; // The opaque parser info describing this device
		public HIDP_CAPS        Caps; // The Capabilities of this hid device.
		public HIDD_ATTRIBUTES  Attributes;

		public byte              []InputReport;
		public short             InputDataLength;

		public byte              []OutputReport;
		public short             OutputDataLength;

		public byte              []FeatureReport;
		public short             FeatureDataLength;
	}
}
